# -*- coding: utf-8 -*- #

module Rouge
  module Lexers
    class ECL < RegexLexer
      title "ECL"
      desc "The ECL Programming Language (hpccsystems.com/about#Platform)"

      tag 'ECL'
      aliases 'ecl'

      filenames '*.ecl'

      mimetypes 'text/x-ecl', 'application/x-ecl'

      def self.keywords
        @keywords ||= Set.new %w(
          AND OR IN NOT ALL ANY AS FROM ATMOST BEFORE BEGINC BEST BETWEEN CASE CONST
          COUNTER CSV DESCEND EMBED ENCRYPT END ENDC ENDEMBED ENDMACRO ENUM EXCEPT
          EXCLUSIVE EXPIRE EXPORT EXTEND FAIL FEW FIRST FLAT FULL FUNCTION FUNCTIONMACRO
          GROUP HEADING HOLE IFBLOCK IMPORT IN JOINED KEEP KEYED LAST LEFT LIMIT LOAD
          LOCAL LOCALE LOOKUP MACRO MANY MAXCOUNT MAXLENGTH MIN SKEW MODULE INTERFACE
          NAMED NOCASE NOROOT NOSCAN NOSORT OF ONLY OPT OUTER OVERWRITE PACKED
          PARTITION PENALTY PHYSICALLENGTH PIPE QUOTE RECORD REPEAT RETURN RIGHT
          ROWS SCAN SELF SEPARATOR SERVICE SHARED SKEW SKIP SQL STORE TERMINATOR
          THOR THRESHOLD TOKEN TRANSFORM TRIM TYPE UNICODEORDER UNSORTED VALIDATE
          VIRTUAL WHOLE WILD WITHIN XML XPATH AFTER CLUSTER COMPRESSED COMPRESSION
          DEFAULT ENCODING ESCAPE FILEPOSITION FORWARD GROUPED INNER INTERNAL
          LINKCOUNTED LITERAL LITTLE_ENDIAN LZW MOFN MULTIPLE NAMESPACE NOCASE
          WNOTRIM NOXPATH ONFAIL OVERWRITE PREFETCH RETRY RIGHT1 RIGHT2 ROWSET
          XPATH SCOPE SMART SOAPACTION STABLE TIMELIMIT TIMEOUT UNORDERED UNSTABLE
          UPDATE USE WIDTH __COMPRESSED__
        )
      end

      def self.builtins
        @builtins ||= Set.new %w(
          ABS ACOS AGGREGATE ALLNODES APPLY ASCII ASIN ASSTRING ATAN ATAN2 AVE CASE
          CATCH CHOOSE CHOOSEN CHOOSESETS CLUSTERSIZE COMBINE CORRELATION COS COSH
          COUNT COVARIANCE CRON DATASET DEDUP DEFINE DENORMALIZE DICTIONARY DISTRIBUTE
          DISTRIBUTED DISTRIBUTION EBCDIC ENTH ERROR EVALUATE EVENT EVENTEXTRA EVENTNAME
          EXISTS EXP FAILCODE FAILMESSAGE FETCH FROMUNICODE FROMXML GETENV GETISVALID GLOBAL
          GRAPH GROUP HASH HASH32 HASH64 HASHCRC HASHMD5 HAVING HTTPCALL HTTPHEADER IF IFF
          INDEX INTFORMAT ISVALID ITERATE JOIN KEYUNICODE LENGTH LIBRARY LIMIT LN LOCAL LOG
          LOOP MAP MATCHED MATCHLENGTH MATCHPOSITION MATCHTEXT MATCHUNICODE MAX MERGE MERGEJOIN
          MIN NOFOLD NOLOCAL NONEMPTY NORMALIZE PARSE PIPE POWER PRELOAD PROCESS PROJECT PULL RANDOM
          RANGE RANK RANKED REALFORMAT RECORDOF REGEXFIND REGEXREPLACE REGROUP REJECTED ROLLUP ROUND
          ROUNDUP ROW ROWDIFF SAMPLE SET SIN SINH SIZEOF SOAPCALL SORT SORTED SQRT STEPPED STORED
          SUM TABLE TAN TANH THISNODE TOPN TOUNICODE TOXML TRANSFER TRANSFORM TRIM TRUNCATE TYPEOF
          UNGROUP UNICODEORDER VARIANCE WHICH WORKUNIT XMLDECODE XMLENCODE XMLTEXT XMLUNICODE
        )
      end

      def self.actions
        @actions ||= Set.new %w(
          APPLY ASSERT BUILD BUILDINDEX EVALUATE FAIL KEYDIFF KEYPATCH LOADXML NOTHOR NOTIFY OUTPUT
          PARALLEL SEQUENTIAL SOAPCALL WAIT
        )
      end

      def self.template
        @template ||= Set.new %w(
          APPEND BREAK DECLARE DEMANGLE END FOR GETDATATYPE IF INMODULE LOOP MANGLE ONWARNING OPTION
          SET STORED UNIQUENAME WORKUNIT
        )
      end

      state :root do
        rule /\s+/m, Text
        rule %r(//(\\.|.)*?$), Literal::String::Regex
        rule %r(/\*), Literal::String::Regex, :multiline_comments
        rule /0x[0-9A-Fa-f]+(?:_[0-9A-Fa-f]+)*/, Num::Integer # hex 0x...
        rule /0b[01]+(?:_[01]+)*/, Num::Integer # binary 0b...
        rule /0[0-9A-Fa-f]+(?:_[0-9A-Fa-f]+)*x/, Num::Integer # hex 0...x
        rule /0[01]+(?:_[01]+)*b/, Num::Integer # binary 0...b
        rule /\d+\.\d+(e[\+\-]?\d+)?/, Num::Integer # real
        rule /0_?[0-7]+(?:_[0-7]+)*/, Num::Integer # octal
        rule /[\d]+(?:_\d+)*/, Num::Integer # integer
        rule /\b(?i:(all|except|export|group|import|keyed|wild|left|right|self|shared|skip|true|false))\b/, Name::Builtin
        rule /\b(?i:(abs|acos|aggregate|allnodes|apply|ascii|asin|asstring|atan|atan2|ave|case|catch|choose|choosen|choosesets|clustersize|combine|correlation|cos|cosh|count|covariance|cron|dataset|dedup|define|denormalize|dictionary|distribute|distributed|distribution|ebcdic|enth|error|evaluate|event|eventextra|eventname|exists|exp|failcode|failmessage|fetch|fromunicode|fromxml|getenv|getisvalid|global|graph|group|hash|hash32|hash64|hashcrc|hashmd5|having|httpcall|httpheader|if|iff|index|intformat|isvalid|iterate|join|keyunicode|length|library|limit|ln|local|log|loop|map|matched|matchlength|matchposition|matchtext|matchunicode|max|merge|mergejoin|min|nofold|nolocal|nonempty|normalize|parse|pipe|power|preload|process|project|pull|random|range|rank|ranked|realformat|recordof|regexfind|regexreplace|regroup|rejected|rollup|round|roundup|row|rowdiff|sample|set|sin|sinh|sizeof|soapcall|sort|sorted|sqrt|stepped|stored|sum|table|tan|tanh|thisnode|topn|tounicode|toxml|transfer|transform|trim|truncate|typeof|ungroup|unicodeorder|variance|which|workunit|xmldecode|xmlencode|xmltext|xmlunicode))\b/, Keyword::Type
        rule /\b(?i:(data|string|qstring|varstring|varunicode|unicode))\d+\b/, Name::Attribute
        rule /\b(?i:(integer|unsigned))[1-8]\b/, Name::Attribute
        rule /\b(?i:(ascii|big_endian|boolean|data|decimal|ebcdic|grouped|integer|linkcounted|pattern|qstring|real|record|rule|set of|streamed|string|token|udecimal|unicode|unsigned|varstring|varunicode))\b/, Name::Attribute
        rule /\b(?i)(u?)decimal(\d+(_\d+)?)\b/, Name::Attribute
        rule /\b(?i:(apply|assert|build|buildindex|evaluate|fail|keydiff|keypatch|loadxml|nothor|notify|output|parallel|sequential|soapcall|wait))\b/, Name::Class
        rule /\b(?i:(import|as|checkpoint|deprecated|failcode|failmessage|failure|global|independent|onwarning|persist|priority|recovery|stored|success|wait|when))\b/, Name::Entity

        rule /'/, Str::Single, :single_string
        rule /"/, Name::Variable, :double_string
        rule /`/, Name::Variable, :backtick

        rule /(#)(?i)(\w+)/, Generic::Subheading

        rule /([A-Za-z_]+)\s*(\()/ do |m|
          if !(self.class.keywords.include? m[1].upcase or self.class.builtins.include? m[1].upcase)
            token Name::Function, m[1]
            token Punctuation, m[2]
          end
          if self.class.keywords.include? m[1].upcase
            token Generic::Traceback
          end
          if self.class.builtins.include? m[1].upcase
            token Name::Builtin
          end
        end
        rule /[\$\w][\w\d]*/ do |m|

        #rule for all remaining tokens assigning token Name for those not being a keyword
        rule /[\$\w][\w\d]*/ do |m|
          if self.class.keywords.include? m[0].upcase
            token Generic::Traceback
          else
            token Name
          end
        end
        
        rule %r([+*/<>=~!@#%^&|?^-]), Operator
        rule /[;:{}()\[\],.]/, Punctuation
      end

      state :multiline_comments do
        rule %r(/[*]), Literal::String::Regex, :multiline_comments
        rule %r([*]/), Literal::String::Regex, :pop!
        rule %r([^/*]+), Literal::String::Regex
        rule %r([/*]), Literal::String::Regex
      end

      state :backtick do
        rule /\\./, Str::Escape
        rule /``/, Str::Escape
        rule /`/, Name::Variable, :pop!
        rule /[^\\`]+/, Name::Variable
      end

      state :single_string do
        rule /\\./, Str::Escape
        rule /''/, Str::Escape
        rule /'/, Str::Single, :pop!
        rule /[^\\']+/, Str::Single
      end

      state :double_string do
        rule /\\./, Str::Escape
        rule /""/, Str::Escape
        rule /"/, Name::Variable, :pop!
        rule /[^\\"]+/, Name::Variable
      end
    end
  end
end
